/*
 *
 * Bubble Demo.
 *
 * This program demonstrates how to perform environment mapping using fixed point.
 * Environment Mapping is a technique used to generate reflections on surfaces.
 * There are several ways to implement Environment Mapping -- this program implements
 * the Dual-Paraboloid approach to Environment Mapping.
 *
 *    For additional background information on Dual-Paraboloid Mapping and a derivation
 *    of the math involved, see the notes located here:
 *
 *    http://developer.nvidia.com/object/dual_paraboloid_env_mapping.html      
 *
 * In addition to performing Environment Mapping, this program demonstrates one way
 * to implement a simple spring physics model in fixed point math.
 *
 * The "bubble" is a dynamic object that you can "poke" and have it deform.
 * The physics for this model are being computed in real time on the CPU.  All of the
 * transformation calculations (for mapping triangles onto the screen) is happening
 * on GoForce.
 *
 * All of the math -- physics, environment mapping, and transformation is happening
 * in real-time.
 *
 * The demo also uses DXT1 compressed textures to ensure that the highest quality
 * textures fit in the smallest possible budget.
 *
 * The demo requires the compressed texture extension "GL_NV_texture_compression_dxt1"
 * and uses the "SGIS_texture_lod" extension to further improve quality.
 *
 * Note:
 * Poking the bubble many times in rapid succession can be fun, but will ultimately
 * cause instability in the bubble mesh.
 *
 */

#include "shape.h"
#include "envcube.h"
#include "quat.h"
#include "inverse.h"
#include <string.h>
#include "nvfxmath/nvfxmath.h"
#include "cursor16x16.h"
#include "genericinput/genericinput.h"
#include <GLES/gl.h>
#include <GLES/egl.h>

//#define DO_FRAMERATE

unsigned char keypress = 255;

static EGLDisplay       dpy;  /* handle to the display          */
static NativeWindowType win;  /* handle to the window           */
static EGLSurface       draw; /* handle to the drawable surface */
static EGLConfig        cfg;  /* handle to an available config  */
static EGLContext       ctx;  /* rendering context              */

#ifdef _WIN32
extern NativeWindowType MyWin32MakeWindow(NativeDisplayType disp,
                                          int x, int y, int w, int h);
#endif

#define M_PI       3.14159265358979323846

enum Mode {MPOINT = 0, MWIRE, MSPHERE, NMODES} mode;

static struct {
    int time_frames;
    int frame_count;
    float startTime;
    float currentTime;
    float duration;
    float elapsed;
    float fps_flag;
    int mouse_flag;
    NvSFx aspect;
    NvSFx near_height;
    NvSFx eye_distance;
    NvSFx near_distance;
    NvSFx far_distance;
	int landscape;
    int mode;
    int texture_flag;
    NvSFx heading;
    NvSFx pitch;
    NvSFx hv;
    NvSFx pv;
    NvSFx screen_x;
    NvSFx screen_y;
    int mouse_x;
    int mouse_y;
    int window_height;
    int window_width; 
    int mouse_down;
    NvSFx orient[4];
    int one_frame;
    int tesslevel;
    int prevpick;
} state;

EnvCube *cube = NULL;
Shape *bubble = NULL;
int lod_clamp_supported;

static GLuint cursor_tid;


#ifdef DO_FRAMERATE
#include <sys/time.h>
double GetTime()
{
    struct timeval tv;
    gettimeofday(&tv, NULL);

    return (double)tv.tv_sec + (double)tv.tv_usec / 1000000.0;
}
#endif


static void BuildCube()
{
    cube = EnvCubeBuildCube();
}

static void BuildBubble( int depth )
{
    if ( bubble )
    {
        ShapeDestroy( bubble );
    }
    switch ( depth )
    {
        case 2:
            bubble = ShapeBuildSphere(9);
            state.tesslevel = 2;
            break;
        case 3:
            bubble = ShapeBuildSphere(13);
            state.tesslevel = 3;
            break;
        case 4:
            bubble = ShapeBuildSphere(25);
            state.tesslevel = 4;
            break;
    }

    ShapeCalcNormals( bubble );
}

static void PassiveMotion(int x, int y)
{
    NvSFx window_x, window_y;
    NvSFx halfh, halfw;

    if ((x > 0) && (x < state.window_width))
        state.mouse_x = x;
    if ((y > 0) && (y < state.window_height))
        state.mouse_y = y;
    
    halfw = (state.window_width >> 1);  /* holds integer width/2  */
    halfh = (state.window_height >> 1); /* holds integer height/2 */
    if ( !state.landscape )
	{
        window_y = (state.window_height - state.mouse_y) - halfh;
        window_y = (window_y << 16); /* convert to NvSFx */
        halfh    = (halfh << 16);    /* convert to NvSFx */
        state.screen_y = NvSFxDiv( window_y, halfh );
        window_x = state.mouse_x - halfw;
        window_x = (window_x << 16); /* convert to NvSFx */
        halfw    = (halfw << 16);    /* convert to NvSFx */
        state.screen_x = NvSFxDiv( window_x, halfw );
	}
	else
	{
        window_y = (state.window_height - state.mouse_y) - halfh;
        window_y = (window_y << 16); /* convert to NvSFx */
        halfh    = (halfh << 16);    /* convert to NvSFx */
        state.screen_x = NvSFxDiv( -window_y, halfh );
        window_x = state.mouse_x - halfw;
        window_x = (window_x << 16); /* convert to NvSFx */
        halfw    = (halfw << 16);    /* convert to NvSFx */
        state.screen_y = NvSFxDiv( window_x, halfw );
	}
}

static void InitBubbleState()
{
    state.time_frames = 2;  // num of frames to time.
    state.frame_count = 0;
#ifdef DO_FRAMERATE
    state.startTime = GetTime();
    state.currentTime = GetTime();
#endif
    state.duration = 60.0;
    state.elapsed = 0.0;
    state.fps_flag = 1;
    state.mouse_flag = 1;

    state.near_height = 0x00010000;   /* 1.0   */
    state.eye_distance = 0x00050000;  /* 5.0   */
    state.near_distance = 0x00020000; /* 2.0   */
    state.far_distance = 0x00280000;  /* 40.0; */
    state.mode = MSPHERE;
    state.texture_flag = 0;
    state.heading = 0x00000000;       /* 0.0   */
    state.pitch = 0x00000000;         /* 0.0   */
    state.hv = 0x00000000;            /* 0.0   */
    state.pv = 0x00000000;            /* 0.0   */
    state.screen_x = 0x00000000;      /* 0.0   */
    state.screen_y = 0x00000000;      /* 0.0   */
    state.mouse_x = 0;
    state.mouse_y = 0;
    state.mouse_down = 0;
    state.one_frame = 0;
    state.prevpick = 0;

#ifndef _WIN32
	BuildBubble(3);
#else
    BuildBubble(4);
#endif
    BuildCube();
#ifndef _WIN32
	state.landscape = 1;
    PassiveMotion( (state.window_width>>1), (state.window_height>>1) - 100 );
#else
	state.landscape = 0;
    PassiveMotion( (state.window_width>>1) - 100, (state.window_height>>1) );
#endif
}

static void InitGraphicsState()
{
    glDepthFunc(GL_LEQUAL);
    glEnable(GL_DEPTH_TEST);
    glClear( GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT );

    /* Initialize the mouse texture. */
    glGenTextures( 1, &cursor_tid );

    glBindTexture( GL_TEXTURE_2D, cursor_tid );
    glTexImage2D( GL_TEXTURE_2D, 0, GL_ALPHA,
                            cursor_res, cursor_res, 0, GL_ALPHA, GL_UNSIGNED_BYTE,
                            cursor_texdata );
    glTexParameterf( GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR );
    glTexParameterf( GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR );
    glTexParameterf( GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP_TO_EDGE );
    glTexParameterf( GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP_TO_EDGE );

}

static void SetAspect()
{
    NvSFx m[4][4];

    state.aspect = 0x00015555;  /* set aspect to 1.33333 */
    glMatrixMode( GL_PROJECTION );
    glLoadIdentity(); 
#ifdef _WIN32
    glFrustumx(NvSFxMul( -state.near_height, state.aspect ),
               NvSFxMul(  state.near_height, state.aspect ),
               -state.near_height,
               state.near_height,  state.near_distance, state.far_distance );
#else
    glFrustumx(-state.near_height,
               state.near_height,
               NvSFxMul( -state.near_height, state.aspect ),
               NvSFxMul(  state.near_height, state.aspect ), state.near_distance, state.far_distance );
#endif
    QuatToMat( m, state.orient );
    // Rotate geometry so that largest dimension is horizontal.
	if ( state.landscape ) glRotatef(-90, 0, 0, 1);

    /* Applying the rotation portion of the viewing transformation. */
    /* Yes, this is actually appended to the projection matrix.     */
    glMultMatrixx( (GLfixed *) &m[0][0]);

    glMatrixMode( GL_MODELVIEW ); 
}

static void DrawMouse(void)
{
    static float texcoords[2*4] = 
    { 
       0.0f, 0.0f,
       0.0f, 1.0f,
       1.0f, 0.0f,
       1.0f, 1.0f 
    };
	GLfloat square[2*4];
	GLfloat halfw;
	GLfloat halfh;
    GLfloat x, y;

    glColor4f( 1.0, 1.0, 1.0, 0.0 );
    glDisable( GL_DEPTH_TEST );

    glMatrixMode( GL_PROJECTION );
    glLoadIdentity(); 
	halfw = (GLfloat)(state.window_width >> 1);
	halfh = (GLfloat)(state.window_height >> 1);
    glOrthof( -halfw, halfw, -halfh, halfh, 0, 1 );

	x = state.mouse_x - halfw;
	y = (state.window_height - state.mouse_y) - halfh;  /* Assumes upper left corner is (0,0) */

    square[2] = x - 8;
    square[3] = y - 8;
    square[0] = x - 8;
    square[1] = y + 8;
    square[4] = x + 8;
    square[5] = y + 8;
    square[6] = x + 8;
    square[7] = y - 8;

    glMatrixMode( GL_MODELVIEW );
    glLoadIdentity(); 

    glEnableClientState ( GL_VERTEX_ARRAY );
	glVertexPointer( 2, GL_FLOAT, 0, square );
	glEnableClientState ( GL_TEXTURE_COORD_ARRAY );
	glTexCoordPointer( 2, GL_FLOAT, 0, texcoords );

    glEnable( GL_TEXTURE_2D );
    glBindTexture( GL_TEXTURE_2D, cursor_tid );

    glTexEnvf( GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_REPLACE );
    glEnable( GL_ALPHA_TEST );
    glAlphaFunc( GL_NOTEQUAL, 0.0f );
    glEnable( GL_BLEND );
    glBlendFunc( GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA );
        glDrawArrays( GL_TRIANGLE_STRIP, 0, 4 );
    glDisable( GL_BLEND );
    glDisable( GL_ALPHA_TEST );

    glDisable( GL_TEXTURE_2D );
    glDisableClientState( GL_VERTEX_ARRAY );
	glDisableClientState( GL_TEXTURE_COORD_ARRAY );

    glEnable( GL_DEPTH_TEST );
}

static void ComputeEyePosition( NvSFx v[3] )
{
    NvSFx pitch_axis[3] = { 0x00010000, 0x00000000, 0x00000000 };
    NvSFx heading_axis[3] = { 0x00000000, 0x00010000, 0x00000000 };

    NvSFx tempQuat1[4];
    NvSFx tempQuat2[4];
    NvSFx m[4][4];
    NvSFx minv[4][4];

    MakeQuat( tempQuat1, state.pitch, pitch_axis );
    MakeQuat( tempQuat2, state.heading, heading_axis );
    MulQuat( state.orient, tempQuat1, tempQuat2 );
    QuatToMat( m, state.orient );

    inverse( minv, m );

    v[0] = NvSFxMul( -minv[2][0], state.eye_distance );
    v[1] = NvSFxMul( -minv[2][1], state.eye_distance );
    v[2] = NvSFxMul( -minv[2][2], state.eye_distance );
}

static void Display(void)
{
    NvSFx EyePos[3];

    SetAspect();

    // Only need to clear the z-buffer
    glClear( GL_DEPTH_BUFFER_BIT );

    glLoadIdentity();
    glScalef(20.0, 20.0, 20.0);

    EnvCubeDraw( cube );

    glLoadIdentity();

    ComputeEyePosition( EyePos );
    glTranslatex( EyePos[0], EyePos[1], EyePos[2] );


    if ( state.texture_flag )
    {
        glScalef(1.3f, 1.3f, 1.3f);
        ShapeDrawSphereMapTexture();
    }
    else
    {
        ShapeCalcVelocity( bubble );
        ShapeFilterVelocity( bubble );
        ShapeCalcNormals( bubble );

        switch( state.mode )
        {
            case MPOINT:
                glColor4f( 1.0, 1.0, 1.0,1.0 );
                ShapeDrawVertices( bubble );
                break;

            case MWIRE:
                glColor4f( 1.0, 1.0, 1.0,1.0 );
                ShapeDrawEdges( bubble );
                break;

            case MSPHERE:
                /* enable dual paraboloid map texturing */
                CalcTexcoordsFromNormal( bubble, EyePos );
                ShapeDraw( bubble );
                break;
        }
    }

	if ( state.mouse_flag )
	{
        DrawMouse();
	}

	eglSwapBuffers( dpy, draw );
}

static void Reshape( int w, int h )
{
    if ( w == 0 || h == 0 ) return;

    state.window_width = w;
    state.window_height = h;
    glViewport( 0, 0, w, h );
}
#ifdef _WIN32
#include <Windows.h>
#endif
static void tick(void)
{
    if ( state.texture_flag )
    {
        /*
        state.hv += -state.heading*0.001f;
        state.pv += -state.pitch*0.001f;
        if ( state.pv < 0.000102207 && state.pv > -0.000102207 )
        {
            state.pv = 0.0;
            state.pitch = 0.0;
        }
        if ( state.hv < 0.000102207 && state.hv > -0.000102207 )
        {
            state.hv = 0.0;
            state.heading = 0.0;
        }
        */
        state.hv = NvSFxAdd( state.hv, NvSFxMul( -state.heading, NvSFxFloat2Fixed( 0.001f ) ) );
        state.pv = NvSFxAdd( state.pv, NvSFxMul( -state.pitch, NvSFxFloat2Fixed( 0.001f ) ) );
    }
    else
    {
        /*
        state.hv += state.screen_x * 0.001f;
        state.pv += state.screen_y * 0.001f;
        state.pv -= state.pitch * 0.0006f;
        */
        state.hv = NvSFxAdd( state.hv, NvSFxMul( state.screen_x, 0x00000041 ) );
        state.pv = NvSFxAdd( state.pv, NvSFxMul( state.screen_y, 0x00000041 ) );
        state.pv = NvSFxSub( state.pv, NvSFxMul( state.pitch, 0x00000027 ) );
    }

    /*
    state.hv *= fix(0.94f);
    state.pv *= fix(0.94f);
    state.heading += state.hv;
    state.pitch += state.pv;
    if (state.heading>M_PI) state.heading -=  (fix) (2.0f*M_PI);
    if (state.heading<-M_PI) state.heading += (fix) (2.0f*M_PI);
    */

    state.hv = NvSFxMul( state.hv, 0x0000f0a3 );
    state.pv = NvSFxMul( state.pv, 0x0000f0a3 );
    state.heading = NvSFxAdd( state.heading, state.hv );
    state.pitch = NvSFxAdd( state.pitch, state.pv );
    if ( state.heading > 0x0003243F ) state.heading = NvSFxSub( state.heading, 0x0006487E );
    if ( state.heading < 0xfffCDBC1 ) state.heading = NvSFxAdd( state.heading, 0x0006487E );

    /* Increment number of frames since last pick. */
    state.prevpick++;

#ifdef DO_FRAMERATE
    state.frame_count++;
    if ( state.frame_count == state.time_frames )
    {
        double time = GetTime();
        state.elapsed = time - state.currentTime;
        state.currentTime = time;
        state.frame_count = 0;
    }

#ifdef DO_TIMELIMIT
    if ( (state.duration > 0.0) &&
         (state.currentTime - state.startTime >= state.duration))
    {
        state.one_frame = 1;
    }
#endif
#endif

#ifdef _WIN32
    //Sleep(100);
#endif
    Display();
}

static void Pick()
{
    NvSFx eye[3];
    NvSFx n[3];
    NvSFx x, y;
    NvSFx m[4][4];
    NvSFx minv[4][4];
    NvSFx norm[3];
    
	/*
	y = state.near_height * state.screen_y;
    x = state.near_height * state.aspect * state.screen_x;

    n[0] = x;
    n[1] = y;
    n[2] = -state.near_distance;
	*/

	y = NvSFxMul( state.near_height, state.screen_y );
    x = NvSFxMul(
		    NvSFxMul( state.near_height, state.aspect ),
	        state.screen_x
	    );

    n[0] = x;
    n[1] = y;
    n[2] = -state.near_distance;

    QuatToMat( m, state.orient );
    inverse( minv, m );

	/*
    eye[0] = minv[2][0]*state.eye_distance;
    eye[1] = minv[2][1]*state.eye_distance;
    eye[2] = minv[2][2]*state.eye_distance;

    norm[0] = n[0]*minv[0][0] + n[1]*minv[1][0] + n[2]*minv[2][0];
    norm[1] = n[0]*minv[0][1] + n[1]*minv[1][1] + n[2]*minv[2][1];
    norm[2] = n[0]*minv[0][2] + n[1]*minv[1][2] + n[2]*minv[2][2];
	*/
	
    eye[0] = NvSFxMul( minv[2][0], state.eye_distance );
    eye[1] = NvSFxMul( minv[2][1], state.eye_distance );
    eye[2] = NvSFxMul( minv[2][2], state.eye_distance );

    norm[0] = NvSFxAdd(
			        NvSFxAdd(
			            NvSFxMul( n[0], minv[0][0] ),
	                    NvSFxMul( n[1], minv[1][0] )
					),
				    NvSFxMul( n[2], minv[2][0] )
			  );
    norm[1] = NvSFxAdd(
			        NvSFxAdd(
			            NvSFxMul( n[0], minv[0][1] ),
	                    NvSFxMul( n[1], minv[1][1] )
					),
				    NvSFxMul( n[2], minv[2][1] )
			  );
    norm[2] = NvSFxAdd(
			        NvSFxAdd(
			            NvSFxMul( n[0], minv[0][2] ),
	                    NvSFxMul( n[1], minv[1][2] )
					),
				    NvSFxMul( n[2], minv[2][2] )
			  );

    ShapePick( bubble, eye, norm );
}

#define MOUSE_DELTA (10)

void keyboard(unsigned char key, int x, int y)
{
#ifndef OLD_KEYSTYLE
    switch (key)
	{
#ifdef _WIN32
        case 38:
        case 104: /* Up Arrow */
			/* Assumes upper left corner is (0,0) */
			if ( state.landscape )
			    PassiveMotion(state.mouse_x+MOUSE_DELTA, state.mouse_y);
			else
			    PassiveMotion(state.mouse_x, state.mouse_y-MOUSE_DELTA);
			return;

        case 40:
        case 98: /* Down Arrow */
			/* Assumes upper left corner is (0,0) */
			if ( state.landscape )
			    PassiveMotion(state.mouse_x-MOUSE_DELTA, state.mouse_y);
			else
			    PassiveMotion(state.mouse_x, state.mouse_y+MOUSE_DELTA);
			break;

        case 37:
        case 100: /* Left Arrow */
			if ( state.landscape )
			    PassiveMotion(state.mouse_x, state.mouse_y-MOUSE_DELTA);
			else
			    PassiveMotion(state.mouse_x-MOUSE_DELTA, state.mouse_y);
			return;

        case 39:
        case 102: /* Right Arrow */
			if ( state.landscape )
			    PassiveMotion(state.mouse_x, state.mouse_y+MOUSE_DELTA);
			else
			    PassiveMotion(state.mouse_x+MOUSE_DELTA, state.mouse_y);
			return;

        case 12:
        case 101: /* '5' key */
            /* Only allow the pick if it has been more than 5 frames since the previous pick. */
            if ( state.prevpick > 5 )
            {
                Pick();
                state.prevpick = 0;
            }
            return;

        case 36:
        case 103: /* '7' <Home> key */
            state.mode = (state.mode+1) % NMODES;
            return;

        case 33:
        case 105: /* '9' <Pg Up> key */
            ShapeCycleNextCubeMap( bubble );
            EnvCycleNextCubeMap( cube );
            return;

        case 35:
        case 97: /* '1' <End> key */
            state.texture_flag = !state.texture_flag;
            return;

        case 34:
        case 99: /* '3' <Pg Dn> key */
            state.mouse_flag = !state.mouse_flag;
            return;

        case 45:
        case 96: /* '0' <Ins> key */
            BuildBubble( state.tesslevel );
            return;

        case 46:
        case 110: /* '.' <Del> key */
            if ( state.tesslevel == 2 )
                BuildBubble(4);
            else if ( state.tesslevel == 3 )
                BuildBubble(2);
            else if ( state.tesslevel == 4 )
                BuildBubble(3);
            return;
#else
        case 56: /* Up Arrow */
			/* Assumes upper left corner is (0,0) */
			if ( state.landscape )
			    PassiveMotion(state.mouse_x+MOUSE_DELTA, state.mouse_y);
			else
			    PassiveMotion(state.mouse_x, state.mouse_y-MOUSE_DELTA);
			return;

        case 50: /* Down Arrow */
			/* Assumes upper left corner is (0,0) */
			if ( state.landscape )
			    PassiveMotion(state.mouse_x-MOUSE_DELTA, state.mouse_y);
			else
			    PassiveMotion(state.mouse_x, state.mouse_y+MOUSE_DELTA);
			break;

        case 52: /* Left Arrow */
			if ( state.landscape )
			    PassiveMotion(state.mouse_x, state.mouse_y-MOUSE_DELTA);
			else
			    PassiveMotion(state.mouse_x-MOUSE_DELTA, state.mouse_y);
			return;

        case 54: /* Right Arrow */
			if ( state.landscape )
			    PassiveMotion(state.mouse_x, state.mouse_y+MOUSE_DELTA);
			else
			    PassiveMotion(state.mouse_x+MOUSE_DELTA, state.mouse_y);
			return;

        case 53: /* '5' key */
            /* Only allow the pick if it has been more than 5 frames since the previous pick. */
            if ( state.prevpick > 5 )
            {
                Pick();
                state.prevpick = 0;
            }
            return;

        case 55: /* '7' <Home> key */
            state.mode = (state.mode+1) % NMODES;
            return;

        case 57: /* '9' <Pg Up> key */
            ShapeCycleNextCubeMap( bubble );
            EnvCycleNextCubeMap( cube );
            return;

        case 49: /* '1' <End> key */
            state.texture_flag = !state.texture_flag;
            return;

        case 51: /* '3' <Pg Dn> key */
            state.mouse_flag = !state.mouse_flag;
            return;

        case 48: /* '0' <Ins> key */
            BuildBubble( state.tesslevel );
            return;

        case 46: /* '.' <Del> key */
            if ( state.tesslevel == 2 )
                BuildBubble(4);
            else if ( state.tesslevel == 3 )
                BuildBubble(2);
            else if ( state.tesslevel == 4 )
                BuildBubble(3);
            return;
		case 27:
            state.one_frame = 1;
			fprintf( stderr, "\n\nEsc key detected... exiting.\n" );
			fprintf( stderr, "If exit was accidental, enable \"Num Lock\" and try again.\n\n" );

			return;

		default:
		fprintf( stderr, "val = %d\n", key );
#endif
    }
#else
    switch (key)
	{
        case '2':
            BuildBubble(2);
            return;

        case '3':
            BuildBubble(3);
            return;

        case '4':
            BuildBubble(4);
            return;

        case 'P':
        case 'p':
            state.mode = MPOINT;
            return;

        case 'W':
        case 'w':
            state.mode = MWIRE;
            return;

        case 'T':
        case 't':
            if ( state.texture_flag )
                state.texture_flag = 0;
            else
                state.texture_flag = 1;
            return;

        case 'S':
        case 's':
            state.mode = MSPHERE;
            return;

        case 'C':
        case 'c':
            ShapeCycleNextCubeMap( bubble );
            EnvCycleNextCubeMap( cube );
            return;

        case 'O':
        case 'o':
            state.landscape = !state.landscape;
            if ( state.landscape )
                PassiveMotion( (state.window_width>>1), (state.window_height>>1) - 50 );
            else
                PassiveMotion( (state.window_width>>1) - 50, (state.window_height>>1) );
            return;

        case 'i':
        case 'I':
			/* Assumes upper left corner is (0,0) */
			if ( state.landscape )
			    PassiveMotion(state.mouse_x+MOUSE_DELTA, state.mouse_y);
			else
			    PassiveMotion(state.mouse_x, state.mouse_y-MOUSE_DELTA);
			return;

        case 'k':
        case 'K':
			/* Assumes upper left corner is (0,0) */
			if ( state.landscape )
			    PassiveMotion(state.mouse_x-MOUSE_DELTA, state.mouse_y);
			else
			    PassiveMotion(state.mouse_x, state.mouse_y+MOUSE_DELTA);
			break;

        case 'j':
        case 'J':
			if ( state.landscape )
			    PassiveMotion(state.mouse_x, state.mouse_y-MOUSE_DELTA);
			else
			    PassiveMotion(state.mouse_x-MOUSE_DELTA, state.mouse_y);
			return;

        case 'l':
        case 'L':
			if ( state.landscape )
			    PassiveMotion(state.mouse_x, state.mouse_y+MOUSE_DELTA);
			else
			    PassiveMotion(state.mouse_x+MOUSE_DELTA, state.mouse_y);
			return;

        case ' ':
                Pick();
            return;

        case 'f':
        case 'F':
            state.fps_flag = !state.fps_flag;
            return;

        case 'm':
        case 'M':
            state.mouse_flag = !state.mouse_flag;
            return;

        case 27:
            state.one_frame = 1;
    }
#endif
}

static void KeyboardInput()
{
	unsigned char c;
	if ( readRawKeyboard(&c) > 0)
	{
	    keyboard( c, 0, 0 );
	}
}


/*
 * CreateGLWindow - open something to draw on
 */
int CreateGLWindow( int width, int height )
{
    EGLint num_cfgs = 0;
    int attrib[] =
    {
        EGL_RED_SIZE,   1, /* At least 1-bit red.   */
        EGL_GREEN_SIZE, 1, /* At least 1-bit green. */
        EGL_BLUE_SIZE,  1, /* At least 1-bit blue.  */
        EGL_DEPTH_SIZE, 1, /* At least 1-bit depth. */
        EGL_NONE
    };

    /* Get a handle to the default display. */
    dpy = eglGetDisplay( EGL_DEFAULT_DISPLAY );
    if (!dpy)
    {
        fprintf( stderr, "Error: Couldn\'t open the EGL default display\n" );
        return( EGL_FALSE );
    }

    /* Initialize the display before calling all other EGL functions. */
    if ( !eglInitialize( dpy, 0, 0 ) )
    {
        fprintf( stderr, "Error: Couldn\'t init the EGL default display\n" );
        return( EGL_FALSE );
    }

    /* Request an EGL config (i.e. pixel format) that has the requested attributes. */
    if ( !eglChooseConfig( dpy, attrib, &cfg, 1, &num_cfgs ) || (num_cfgs < 1) )
    {
        fprintf( stderr, "Error: Couldn\'t find a config on the EGL default display\n" );
        return( EGL_FALSE );
    }

#ifdef _WIN32
    // Use this function to create your window at location: (391,90).
    // This function will create a borderless, top-level window with
    // a 240x320 resolution.
    win = MyWin32MakeWindow(EGL_DEFAULT_DISPLAY, 351, 150, 320, 240);
#else
    width = eglDisplayWidthNV( dpy, cfg );
    height = eglDisplayHeightNV( dpy, cfg );
    win = eglMakeWindowNV( EGL_DEFAULT_DISPLAY, 0, 0, width, height );
#endif
    if (!win)
	{
        fprintf( stderr, "Error: Couldn\'t create window\n" );
        return( EGL_FALSE );
    }

    /* Create the drawable surface and rendering context. */
    draw = eglCreateWindowSurface( dpy, cfg, win, 0 );
    ctx = eglCreateContext( dpy, cfg, 0, 0 );

    /* Make the context current and bind the newly created drawable */
    /* surface as the current "draw" and "read" surface.            */
    eglMakeCurrent( dpy, draw, draw, ctx );

#ifdef _WIN32
    Reshape( 320, 240 );
#else
    Reshape( width, height );
#endif

    return( EGL_TRUE );
}

/*
 * KillGLWindow - shut down rendering
 */
void KillGLWindow()
{
    if (dpy)
    {
        if (ctx) eglDestroyContext( dpy, ctx );
        if (draw) eglDestroySurface( dpy, draw );
        if (win) eglDestroyWindowNV( win );
        eglTerminate( dpy );
    }
}

static GLboolean QueryExtension(char *extName)
{
    int n;
    char *p = (char *) glGetString(GL_EXTENSIONS);
    char *end = p + strlen(p); 
    while ( p < end )
    {
        n = (int) strcspn(p, " ");
        if ((strlen(extName)==n) && (strncmp(extName,p,n)==0)) {
            return GL_TRUE;
        }
        p += (n + 1);
    }
    return GL_FALSE;
}

//#ifdef _WIN32
//INT WINAPI WinMain(HINSTANCE hInst, HINSTANCE hPrevInstance, LPSTR lpCmdLine, INT nShowCmd )
//#else
int main(int argc, char **argv)
//#endif
{
    if ( !CreateGLWindow( 240, 320 ) )
    //if ( !CreateGLWindow( 320, 240 ) )
    {
        fprintf( stderr, "Display initialization failed.\n" );
        fprintf( stderr, "Exiting.\n" );
        exit( -1 );
    }

    /* Now that the context is created, verify the extension is supported. */
    if ( !QueryExtension( "GL_NV_texture_compression_dxt1" ) )
    {
        fprintf( stderr, "Extension: GL_NV_texture_compression_dxt1 not supported.\n" );
        fprintf( stderr, "It is required to run this demo.  Exiting.\n" );
        exit( -1 );
    }

    /* We will clamp the lod of the paraboloid textures if this extension is supported. */
    lod_clamp_supported = QueryExtension( "GL_SGIS_texture_lod" );
    if ( !lod_clamp_supported )
    {
        fprintf( stderr, "Extension: GL_SGIS_texture_lod is not supported.\n" );
        fprintf( stderr, "The demo will run but at reduced quality.\n" );
        fprintf( stderr, "You may notice some \"sparkles\" at silhouette edges.\n\n" );
    }

    InitGraphicsState();
    InitBubbleState();

#ifndef _WIN32
	initRawKeyboard();
#endif

    /* Print Usage */
    fprintf( stderr, "\n" );
    fprintf( stderr, "Bubble Keyboard Controls:\n" );
    fprintf( stderr, "------------------------\n" );
    fprintf( stderr, "    keypad 8 -- Move mouse cursor up\n" );
    fprintf( stderr, "    keypad 2 -- Move mouse cursor down\n" );
    fprintf( stderr, "    keypad 4 -- Move mouse cursor left\n" );
    fprintf( stderr, "    keypad 6 -- Move mouse cursor right\n" );
    fprintf( stderr, "    keypad 5 -- Poke bubble\n" );
    fprintf( stderr, "    keypad 0 -- Reset bubble\n" );
    fprintf( stderr, "    keypad 7 -- Cycle solid, point, wireframe\n" );
    fprintf( stderr, "    keypad 9 -- Toggle environment map\n" );
    fprintf( stderr, "    keypad 1 -- Display dual-paraboloid texture\n" );
    fprintf( stderr, "    keypad 3 -- Toggle hide/show mouse cursor\n" );
    fprintf( stderr, "    keypad . -- Cycle bubble tessellation levels\n" );
    fprintf( stderr, "    <ESC>    -- Quit\n" );

    do {
#ifndef _WIN32
		KeyboardInput();
#else
        keyboard( keypress, 0, 0 );
        keypress = 255;
#endif

#ifdef DO_FRAMERATE
        fprintf( stderr, "Drawing frame #%d\n", state.frame_count );
        if (state.fps_flag && state.elapsed != 0.0)
        {
            float seconds = float(state.elapsed);
            float fps = state.time_frames / seconds;
            fprintf( stderr, "fps = %.2f\n", fps );
        }
#endif
        tick();
    } while ( !state.one_frame );

#ifndef _WIN32
	closeRawKeyboard();
#endif

    KillGLWindow();

#ifdef DO_FRAMERATE
#ifdef DO_TIMELIMIT
    if ( state.duration != 0.0f )
    {
        printf("Exiting after %f seconds\n", state.duration);
    }
#endif
#endif

    exit(0);
}
