/*
 * Simple app that demonstrates how to create a rendering context
 * and draw a triangle using fixed point entry points.
 */

#include <GLES/egl.h>
#include <GLES/gl.h>
#include <stdlib.h>
#include <stdio.h>
#include <stdarg.h>

static EGLDisplay       dpy;  /* handle to the display          */
static NativeWindowType win;  /* handle to the window           */
static EGLSurface       draw; /* handle to the drawable surface */
static EGLConfig        cfg;  /* handle to an available config  */
static EGLContext       ctx;  /* rendering context              */

/*
 * CreateGLWindow - open something to draw on
 */
int CreateGLWindow( int width, int height )
{
    EGLint num_cfgs = 0;
    int attrib[] =
    {
        EGL_RED_SIZE,   1, /* At least 1-bit red.   */
        EGL_GREEN_SIZE, 1, /* At least 1-bit green. */
        EGL_BLUE_SIZE,  1, /* At least 1-bit blue.  */
        EGL_DEPTH_SIZE, 1, /* At least 1-bit depth. */
        EGL_NONE
    };

    /* Get a handle to the default display. */
    dpy = eglGetDisplay( EGL_DEFAULT_DISPLAY );
    if (!dpy)
    {
        fprintf( stderr, "Error: Couldn\'t open the EGL default display\n" );
        return( EGL_FALSE );
    }

    /* Initialize the display before calling all other EGL functions. */
    if ( !eglInitialize( dpy, 0, 0 ) )
    {
        fprintf( stderr, "Error: Couldn\'t init the EGL default display\n" );
        return( EGL_FALSE );
    }

    /* Request an EGL config (i.e. pixel format) that has the requested attributes. */
    if (!eglChooseConfig( dpy, attrib, &cfg, 1, &num_cfgs ) || (num_cfgs < 1))
    {
        fprintf( stderr, "Error: Couldn\'t find a config on the EGL default display\n" );
        return( EGL_FALSE );
    }

    /* To create a full screen window, query display width and height */
    /* and use them as parameters when making the window.             */
#ifndef _WIN32
    width = eglDisplayWidthNV( dpy, cfg );
    height = eglDisplayHeightNV( dpy, cfg );
#endif
    win = eglMakeWindowNV( EGL_DEFAULT_DISPLAY, 0, 0, width, height );
    if (!win)
	{
        fprintf( stderr, "Error: Couldn\'t create window\n" );
        return( EGL_FALSE );
    }

    /* Create the drawable surface and rendering context. */
    draw = eglCreateWindowSurface( dpy, cfg, win, 0 );
    ctx = eglCreateContext( dpy, cfg, 0, 0 );

    /* Make the context current and bind the newly created drawable */
    /* surface as the current "draw" and "read" surface.            */
    eglMakeCurrent( dpy, draw, draw, ctx );

    return( EGL_TRUE );
}

/*
 * KillGLWindow - shut down rendering
 */
void KillGLWindow()
{
    if (dpy)
    {
        /* Delete created resources. */
        if (ctx) eglDestroyContext( dpy, ctx );
        if (draw) eglDestroySurface( dpy, draw );
        if (win) eglDestroyWindowNV( win );
        
        /* Release resources associated with the display. */
        eglTerminate( dpy );
    }
}

static void CleanUp()
{
	KillGLWindow();
}

#ifndef _WIN32
static void FatalError(const char *msg, ...)
     __attribute__ ((format (printf, 1, 2), noreturn));
#else
DECLSPEC_NORETURN static void FatalError(const char *msg, ...);
#endif

static void FatalError(const char *msg, ...)
{
    va_list ap;
#ifdef _WIN32
    char error[512];
#endif
    va_start(ap, msg);
    vfprintf(stderr, msg, ap);
    va_end(ap);
    fputc('\n', stderr);
#ifdef _WIN32
   _vsnprintf( error, sizeof(error), msg, ap ); 
    MessageBox( NULL, error, "FatalError Message", MB_OK );
#endif
    CleanUp();
	exit(-1);
}

static void CheckError(const char *msg)
{
    int e = glGetError();
    if (e) FatalError("Error %s: %04x", msg, e);
}

/* OpenGL ES supports 16.16 fixed-point interfaces.
 *
 * 16.16 fixed-point representation is defined in the OpenGL ES spec as
 * a 32-bit two's complement number with 16 bits of fraction.  Based on
 * this definition, floating-point and integer values may be converted
 * to fixed-point by taking the value multiplying it by 2^16 then converting
 * the resulting value to an integer.
 *
 * Using this approach:
 *     1.0 is 0x00010000 in fixed point
 *    -1.0 is 0xFFFF0000 in fixed point
 *    -0.5 is 0x00008000 in fixed point
 *    -0.5 is 0xFFFF8000 in fixed point
 *
 * When using fixed point representations of numbers, the rules for
 * multiplication and division are slightly different.  For this reason
 * when two fixed-point numbers are stored in variables of type 'int' it is
 * generally inappropriate to use the '/' and '*' operators when operating
 * on them.
 * 
 * Note: In the code below, "ROOT_3_OVER_2/3" produces the intended result
 * because the denominator is actually the integer value 3 -- not the 16.16
 * fixed- point representation of the value 3.
 *
 * For a general purpose math library, consider using the nvfxmath library
 * included with the NVIDIA Handheld SDK.  This library defines several
 * math functions for operating on 16.16 fixed-point types.
 */

#define FIXINT(x) (x << 16)
#define FIXHALF (FIXINT(1) >> 1)

#define ROOT_3_OVER_2 ((GLfixed) (unsigned) (0.8660254 * (1 << 16)))
#define ROOT_3_OVER_6 (ROOT_3_OVER_2/3)

static GLubyte ind [3] = { 0, 1, 2 };
static GLfixed vert[6] = { FIXHALF, -ROOT_3_OVER_6,
                          -FIXHALF, -ROOT_3_OVER_6,
                           FIXINT(0),  ROOT_3_OVER_2 - ROOT_3_OVER_6 };
static GLfixed tex [6] = { FIXINT(1), FIXINT(0),
                           FIXINT(0), FIXINT(0),
                           FIXHALF, FIXINT(1) };
static GLfixed col [9] = { FIXINT(1), FIXINT(0), FIXINT(0),
                           FIXINT(0), FIXINT(1), FIXINT(0),
                           FIXINT(0), FIXINT(0), FIXINT(1) };

static void InitGraphicsState()
{
    glClearColorx(0, 0, 0, 0);
    glOrthox(-FIXINT(1), FIXINT(1), -FIXINT(1),
             FIXINT(1), -FIXINT(1), FIXINT(1));
    glDepthFunc(GL_LEQUAL);
    glEnable(GL_DEPTH_TEST);
    glEnableClientState(GL_VERTEX_ARRAY);
    glEnableClientState(GL_TEXTURE_COORD_ARRAY);
    glEnableClientState(GL_COLOR_ARRAY);
    glVertexPointer  (2, GL_FIXED, 0, (void *) vert);
    glTexCoordPointer(2, GL_FIXED, 0, (void *) tex);
    glColorPointer   (3, GL_FIXED, 0, (void *) col);

    glClear( GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT );

	CheckError( "InitGraphicsState" );
}

static void Display(void)
{
    glClear( GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT );
    glLoadIdentity();
    glDrawElements(GL_TRIANGLES, 3, GL_UNSIGNED_BYTE, ind);
    eglSwapBuffers( dpy, draw );
}

static void tick(void)
{
    Display();
}

int main()
{
	if ( !CreateGLWindow( 240, 320 ) )
	{
		FatalError( "Display initialization failed.\nExiting." );
	}

    InitGraphicsState();

    while ( 1 )
    {
        tick();
    }

    CleanUp();

	return 0;
}

