/*
 * nvfxmath.h
 *
 * Optimized S15.16 fixed point math library routines.
 */

#if !defined(NVFXMATH_H)
#define NVFXMATH_H

#include <stdlib.h>  /* for labs() prototype */

#ifdef __cplusplus
extern "C" {
#endif

typedef int NvSFx;
// TODO: support for unsigned 16.16?

#define NvSFxAbs(A)   (labs((NvSFx)(A)))
#define NvSFxFloor(A) ((((NvSFx)(A)) >> 16) << 16)

/* Result is undefined if sum > 7FFF.FFFF or sum < 8000.0000 */
#define NvSFxAdd(A,B) (((NvSFx)(A)) + ((NvSFx)(B)))

/* Result is undefined if difference > 7FFF.FFFF or difference < 8000.0000 */
#define NvSFxSub(A,B) (((NvSFx)(A)) - ((NvSFx)(B)))

/* Result is undefined if product > 7FFF.FFFF or product < 8000.0000 */
#define NvSFxMul(A,B) ((NvSFx)((((long long)(NvSFx)(A)) * ((NvSFx)(B))) >> 16))

/* Result undefined if input > 7FFF.0000 */
#define NvSFxCeil(A)  (-NvSFxFloor(-((NvSFx)(A))))

/* 
 * NvSFxPow approximates A^B (where "^" denotes "raised to the power of"). 
 * In general, the result is undefined if A is not a non-zero, non-negative 
 * number, i.e. A > 0000.0000. However pow(0,0) is 0001.0000 as demanded 
 * by OpenGL-ES. The result is undefined if pow(A,B) > 7FFF.FFFF, The result
 * is also undefined if log2(A)*B is not in range [8000.0000, 7FFF.FFFF].
 *
 * The relative error of the approximation is about 3*10^-5. In other words, 
 * worst case approximation error increases with the magnitude of the result, 
 * and for large results, no more than the most significant 16 bits of the 
 * result can be guaranteed to be correct (although many results will have 
 * better accuracy.
 */
#define NvSFxPow(A,B) (NvSFxPow2(NvSFxMul((B),NvSFxLog2(A))))

#if defined(ARM) || defined(XSCALE)
/* 
 * Division is exact and the result is equal to the quotient of dividend and
 * divisor rounded towards 0, i.e. truncated after 16 fraction bits. The 
 * quotient is undefined if abs(dividend) > (7FFF.FFFF * abs(divisor)).
 */
NvSFx NvSFxDiv(NvSFx dividend, NvSFx divisor);

/* 
 * Reciprocal is exact and the result is equal to the quotient of 1 / divisor 
 * rounded towards 0, i.e. truncated after 16 fraction bits. The quotient is 
 * undefined if abs(dividend) > (7FFF.FFFF * abs(divisor)).
 */
NvSFx NvSFxRecip(NvSFx divisor);

/*
 * NvSFxRemainder is the fixed-point equivalent of the ANSI-C function fmod().
 * It computes dividend - trunc(dividend / divisor) * divisor, where trunc(x)
 * = (x >= 0.0) ? floor(x) : ceil(x)). The remainder has the same sign as the
 * dividend. The result is undefined if 
 * abs(dividend) > (7FFF.FFFF * abs(divisor)).
 */
NvSFx NvSFxRemainder(NvSFx dividend, NvSFx divisor);
#else
#define NvSFxDiv(A,B) ((NvSFx)((((long long)(NvSFx)(A)) << 16) / ((NvSFx)(B))))
#define NvSFxRecip(B) ((NvSFx)((((long long)(0x10000)) << 16) / ((NvSFx)(B))))
#define NvSFxRemainder(A,B) ((NvSFx)((((long long)(NvSFx)(A)) << 16) % ((NvSFx)(B))))
#endif

/* 
 * NvSFxCosD computes an approximation to the cosine of deg, where deg is an
 * angle given in degrees (a full circle having 360 degrees). The input must
 * be >= 0 and less than 360 degress, i.e. 0000.0000 <= deg <= 0167.FFFF. The
 * function returns 0 for arguments outside this range. The magnitude of the
 * approximation error does not exceed 4/65536 over the entire supported 
 * argument range.
 */ 
NvSFx NvSFxCosD(NvSFx deg);

/* 
 * NvSFxSinD computes an approximation to the sine of deg, where deg is an
 * angle given in degrees (a full circle having 360 degrees). The input must
 * be >= 0 and less than 360 degress, i.e. 0000.0000 <= deg <= 0167.FFFF. The
 * function returns 0 for arguments outside this range. The magnitude of the
 * approximation error does not exceed 4/65536 over the entire supported 
 * argument range.
 */
NvSFx NvSFxSinD(NvSFx deg);

/*
 * Square root is exact and teh result is equal to sqrt(radicand) rounded
 * towards 0, i.e. truncated after 16 fraction bits. The result is undefined
 * if radicand does not satisfy 0000.0000 <= radicand < 7FFF.FFFF, i.e. the
 * radicand must be greater than or equal to zero.
 */
NvSFx NvSFxSqrt(NvSFx radicand);

/*
 * NvSFxRecipSqrt computes an approximation to 1/sqrt(radicand). The result is
 * undefined if radicand does not satisfy 0000.0000 < radicand < 7FFF.FFF, i.e
 * the argument must be a non-zero non-negative number. For all but 161 of the
 * in-range arguments, the result differs from a reference function (floating-
 * point converted to fixed-point) by at most +/- 1/65536. Slightly larger 
 * error occurs for a few arguments less than 0000.07AE (about 0.03). The 
 * magnitude of the error does not exceed 7/65536 over the entire supported 
 * argument range.
 */
NvSFx NvSFxRecipSqrt(NvSFx radicand);

/*
 * NvSFxLog2 computes an approximation to the logarithm base 2 of x. The result
 * is undefined if the argument x does not satisfy 0000.0000 < x < 7FFF.FFFF,
 * i.e. argument must be a non-zero non-negative number. Over the supported
 * argument range 99.998% of the results either match the reference function 
 * (floating-point rounded to S15.16) or are off by -1/65536. Only 0.002% are 
 * off by -2/65536. No larger errors were observed in an exhaustive test.
 */
NvSFx NvSFxLog2(NvSFx x);

/*
 * NvSFxPow2 computes an approximation to 2^x. The result is undefined if the 
 * argument x is > 000E.FFFF (about 14.99998). The approximation provides 
 * about 22.5 bits of accuracy and the result is not off by more than -2/65536
 * (as compared to a floating-point reference converted to S15.16) if the 
 * argument is less than 0007.8000. The approximation error increases for 
 * larger arguments, but does not exceed 180/65536 over the entire supported 
 * argument range.
 */
NvSFx NvSFxPow2(NvSFx x);

/*
 * NvSFxFloat2Fixed converts the floating-point number f into a S15.16 fixed
 * point number. The result is identical to ((int)(f * 65536.0f)), but using
 * this function avoids floating-point computation. The result is undefined
 * if f does not satisfy -32768.0 < f < 32767.9995.
 */
NvSFx NvSFxFloat2Fixed(float f);

/*
 * NvSFxAtan2D computes an approximation to the principal value of arctangent 
 * of y/x, using the signs of both arguments to determine the quadrant of the
 * return value. The value of the arctangent returned is in degrees (where a
 * full circle comprises 360 degrees) and is in [-180, 180]. Zero is returned 
 * if both arguments are zero. The magnitude of the error should not exceed 
 * 2/65536 (determined by comparing a large number of results to a floating-
 * point reference).
 */
NvSFx NvSFxAtan2D (NvSFx y, NvSFx x);

/*
 * NvSFxAsinD computes an approximation to the principal value of arcsine of 
 * x. The value of the arc sine returned is in degrees (where a full circle 
 * comprises 360 degrees) and is in [-90, 90]. The result is undefined if the 
 * input does not satisfy -1 <= x <= 1. The magnitude of the approximation 
 * error does not exceed 3/65536 over the entire legal argument range.
 */
NvSFx NvSFxAsinD (NvSFx x);

/*
 * NvSFxAcosD computes an approximation to the principal value of arcsine of 
 * x. The value of the arc cosine returned is in degrees (where a full circle 
 * comprises 360 degrees) and is in [-90, 90]. The result is undefined if the 
 * input does not satisfy -1 <= x <= 1. The magnitude of the approximation 
 * error does not exceed 3/65536 over the entire legal argument range.
 */
NvSFx NvSFxAcosD (NvSFx x);

/* full range versions of sine and cosine */
NvSFx NvSFxCosDFullRange (NvSFx deg);
NvSFx NvSFxSinDFullRange (NvSFx deg);

#ifdef __cplusplus
}
#endif

#endif /* NVFXMATH_H */
